"""# Model comparison tests"""

from .metrics import check_scoring

import numpy as np
import pandas as pd
from scipy.stats import t as t_distribution
from sklearn.base import clone, is_classifier
from sklearn.model_selection import check_cv
from joblib import Parallel, delayed

def _compute_pairwise_diff(
        estimators, X, y, repetitions, cv, scoring=None, n_jobs=None
    ):
    """
    Returns
    -------
    pairwise_diff : np.array of shape (n_pairs, r*k)
        Array of pairwise differences between estimator scores across 
        repetitions and folds.
    """
    def fold_score(estimator, train_idx, test_idx):
        if isinstance(X, pd.DataFrame):
            X_train, X_test = X.iloc[train_idx], X.iloc[test_idx]
        else:
            X_train, X_test = X[train_idx], X[test_idx]
        y_train, y_test = y[train_idx], y[test_idx]
        estimator.fit(X_train, y_train)
        return scoring(estimator, X_test, y_test)
    
    scoring = check_scoring(scoring, classifier=is_classifier(estimators[0]))
    cv = check_cv(cv, y=y, classifier=is_classifier(estimators[0]))
    if hasattr(cv, 'shuffle'):
        cv.shuffle = True
    scores = Parallel(n_jobs=n_jobs)(
        delayed(fold_score)(
            clone(est), train_idx, test_idx
        )
        for _ in range(repetitions)
        for train_idx, test_idx in cv.split(X)
        for name, est in estimators
    )
    scores = np.array(scores).reshape(-1, len(estimators))
    return np.array([
        scores[:,i]-scores[:,j]
        for i in range(len(estimators)) 
        for j in range(i+1, len(estimators))
    ]).T

def _collect_dataframe(
        estimators, pairwise_diff_mean, pairwise_diff_std, t_stats, p_vals
    ):
    # collects test results into a pandas dataframe
    df = pd.DataFrame([
        {
            'Estimator1': estimators[i][0],
            'Estimator2': estimators[j][0]
        }
        for i in range(len(estimators))
        for j in range(i+1, len(estimators))
    ])
    df['PerformanceDifference'] = pairwise_diff_mean
    df['Std'] = pairwise_diff_std
    df['t-stat'] = t_stats
    df['p-value'] = p_vals
    return df

def corrected_repeated_kfold_cv_test(
        estimators, X, y, repetitions=10, cv=10, scoring=None, n_jobs=None
    ):
    """
    Performs pairwise corrected repeated k-fold cross-validation tests. See [Bouckaert and Frank](https://www.cs.waikato.ac.nz/~eibe/pubs/bouckaert_and_frank.pdf).

    Parameters
    ----------
    estimators : list
        List of (name, estimator) tuples.

    X : array-like of shape (n_samples, n_features)
        Features.

    y : array-like of shape (n_samples, n_targets)
        Targets.

    repetitions : int, default=10
        Number of cross-validation repetitions.

    cv : int, cross-validation generator, or iterable, default=10
        Scikit-learn style cv parameter.

    scoring : str, callable, list, tuple, or dict, default=None
        Scikit-learn style scoring parameter.

    n_jobs : int, default=None
        Number of jobs to run in parallel.

    Returns
    -------
    results_df : pd.DataFrame

    Examples
    --------
    ```python
    from fast_automl.test import corrected_repeated_kfold_cv_test

    from sklearn.datasets import load_boston
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import Ridge
    from sklearn.svm import SVR

    X, y = load_boston(return_X_y=True)
    corrected_repeated_kfold_cv_test(
    \    [
    \        ('rf', RandomForestRegressor()),
    \        ('ridge', Ridge()),
    \        ('svm', SVR())
    \    ],
    \    X, y, n_jobs=-1
    )
    ```

    Out:

    ```
    Estimator1 Estimator2  PerformanceDifference       Std     t-stat       p-value
    \        rf      ridge               0.165030  0.030266   5.452600  3.652601e-07   
    \        rf        svm               0.670975  0.045753  14.665154  1.460994e-26  
    \     ridge        svm               0.505945  0.045031  11.235469  2.258586e-19
    ```
    """
    cv = check_cv(cv, y=y, classifier=is_classifier(estimators[0]))
    pairwise_diff = _compute_pairwise_diff(
        estimators, X, y, repetitions, cv, scoring, n_jobs
    )
    # Nadeau and Bengio correction
    # https://www.cs.waikato.ac.nz/~eibe/pubs/bouckaert_and_frank.pdf
    pairwise_diff_var = pairwise_diff.var(axis=0, ddof=1)
    k = cv.get_n_splits(X)
    pairwise_diff_var *= 1./(k * repetitions) + 1./(k-1)
    # compute statistics
    pairwise_diff_mean = pairwise_diff.mean(axis=0)
    pairwise_diff_std = np.sqrt(pairwise_diff_var)
    t_stats = pairwise_diff_mean / pairwise_diff_std
    df = k * repetitions - 1
    p_vals = [t_distribution.sf(abs(t_stat), df)*2 for t_stat in t_stats]
    return _collect_dataframe(estimators, pairwise_diff_mean, pairwise_diff_std, t_stats, p_vals)

def r_by_k_cv_test(
        estimators, X, y, repetitions=5, cv=2, scoring=None, n_jobs=None
    ):
    """
    Performs pariwise RxK (usually 5x2) cross-validation tests. See [here](https://www.kaggle.com/ogrellier/parameter-tuning-5-x-2-fold-cv-statistical-test).

    Parameters
    ----------
    estimators : list
        List of (name, estimator) tuples.

    X : array-like of shape (n_samples, n_features)
        Features.

    y : array-like of shape (n_samples, n_targets)
        Targets.

    repetitions : int, default=10
        Number of cross-validation repetitions.

    cv : int, cross-validation generator, or iterable, default=10
        Scikit-learn style cv parameter.

    scoring : str, callable, list, tuple, or dict, default=None
        Scikit-learn style scoring parameter.

    n_jobs : int, default=None
        Number of jobs to run in parallel.

    Returns
    -------
    results_df : pd.DataFrame

    Examples
    --------
    ```python
    from fast_automl.test import r_by_k_cv_test

    from sklearn.datasets import load_boston
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import Ridge
    from sklearn.svm import SVR

    X, y = load_boston(return_X_y=True)
    r_by_k_cv_test(
    \    [
    \        ('rf', RandomForestRegressor()),
    \        ('ridge', Ridge()),
    \        ('svm', SVR())
    \    ],
    \    X, y, n_jobs=-1
    )
    ```

    Out:

    ```
    Estimator1 Estimator2  PerformanceDifference       Std     t-stat   p-value
    \        rf      ridge               0.143314  0.026026   5.506631  0.002701
    \        rf        svm               0.659547  0.035824  18.410644  0.000009
    \     ridge        svm               0.516233  0.021601  23.898480  0.000002
    ```
    """
    cv = check_cv(cv, y=y, classifier=is_classifier(estimators[0]))
    pairwise_diff = _compute_pairwise_diff(
        estimators, X, y, repetitions, cv, scoring, n_jobs
    )
    # compute variance of performance difference across folds
    n_pairs = int(.5 * len(estimators) * (len(estimators)-1))
    diff = pairwise_diff.reshape(cv.get_n_splits(X), repetitions, n_pairs)
    pairwise_diff_var = diff.var(axis=0, ddof=1).mean(axis=0)
    # compute statistics
    pairwise_diff_mean = pairwise_diff[0]
    pairwise_diff_std = np.sqrt(pairwise_diff_var)
    t_stats = pairwise_diff_mean / pairwise_diff_std
    df = repetitions
    p_vals = [t_distribution.sf(abs(t_stat), df)*2 for t_stat in t_stats]
    return _collect_dataframe(
        estimators, pairwise_diff_mean, pairwise_diff_std, t_stats, p_vals
    )


"""# Cross-validation estimators

Examples
--------
```python
from fast_automl.cv_estimators import RandomForestClassifierCV

from sklearn.datasets import load_digits
from sklearn.model_selection import cross_val_score, train_test_split

X, y = load_digits(return_X_y=True)
X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, shuffle=True)
clf = RandomForestClassifierCV().fit(X_train, y_train, n_jobs=-1)
print('Cross val score: {:.4f}'.format(cross_val_score(clf.best_estimator_, X_train, y_train).mean()))
print('Test score: {:.4f}'.format(clf.score(X_test, y_test)))
```

Out:

```
Cross val score: 0.9696
Test score: 0.9800
```
"""

from .metrics import check_scoring

from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, is_classifier
from sklearn.decomposition import PCA
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor, RandomForestClassifier, RandomForestRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import ElasticNet, LassoLars, LogisticRegression, Ridge
from sklearn.metrics import make_scorer, r2_score
from sklearn.model_selection import RandomizedSearchCV
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures, StandardScaler
from sklearn.svm import SVC, SVR
from scipy.stats import expon, geom, uniform, poisson, randint
from xgboost import XGBClassifier, XGBRegressor

from sklearn.utils._testing import ignore_warnings
from sklearn.exceptions import ConvergenceWarning


class CVBaseEstimator(BaseEstimator):
    """
    Base class for all CV estimators.

    Parameters
    ----------
    preprocessors : list, default=[]
        Preprocessing steps.

    param_distributions : dict, default={}
        Maps names of parameters to distributions. This overrides 
        parameters returned by the `get_param_distributions` method.

    Attributes
    ----------
    best_estimator_ : estimator
        Estimator which attained the best CV score under randomized search.

    best_score_ : scalar
        Best CV score attained by any estimator.

    cv_results_ : list
        List of (mean CV score, parameters) tuples.
    """
    def __init__(self, preprocessors=[], param_distributions={}):
        self.preprocessors = (
            preprocessors if isinstance(preprocessors, list) 
            else [preprocessors]
        )
        self.param_distributions = param_distributions

    def get_param_distributions(self, param_distributions={}):
        """
        Parameters
        ----------
        param_distributions : dict, default={}
            These are overridden by the `param_distributions` parameter passed
            to the constructor.

        Returns
        -------
        param_distributions : dict
            Parameter distributions used for randomized search.
        """
        param_distributions.update(self.param_distributions)
        return param_distributions

    def make_estimator(self, **params):
        raise NotImplementedError('make_estimator method not implemented')

    @ignore_warnings(category=ConvergenceWarning)
    def fit(self, X, y, n_iter=10, n_jobs=None, scoring=None):
        """
        Fits a CV estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        n_iter : int, default=10
            Number of iterations to use in randomized search.

        n_jobs : int or None, default=None
            Number of background jobs to use in randomized search.

        scoring : str or callable, default=None
            A str (see model evaluation documentation) or a scorer callable 
            object / function with signature scorer(estimator, X, y) which 
            should return only a single value. If `None`, the estimator's 
            default `score` method is used.
        """
        scoring = check_scoring(scoring, classifier=is_classifier(self))
        est = RandomizedSearchCV(
            self.make_estimator(),
            self.get_param_distributions(X, y),
            n_iter=n_iter,
            n_jobs=n_jobs,
            scoring=scoring
        ).fit(X, y)
        self.cv_results_ = list(zip(
            est.cv_results_['mean_test_score'], est.cv_results_['params']
        ))
        self.best_estimator_ = est.best_estimator_
        self.best_score_ = est.best_score_
        return self

    def predict(self, X):
        """
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples, n_targets)
            Predicted values
        """
        assert hasattr(self, 'best_estimator_'), (
            'Estimator has not been fitted. Call `fit` before `predict`'
        )
        return self.best_estimator_.predict(X)

    def predict_proba(self, X):
        """
        Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the sample for each classes on the model, ordered 
            by `self.classes_`.

        Notes
        -----
        Only applicable for classifiers.
        """
        assert hasattr(self, 'best_estimator_'), (
            'Estimator has not been fitted. Call `fit` before `predict`'
        )
        return self.best_estimator_.predict_proba(X)


class RandomForestClassifierCV(ClassifierMixin, CVBaseEstimator):    
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            RandomForestClassifier()
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'randomforestclassifier__n_estimators': geom(1./100)
        })


class PCARandomForestClassifierCV(ClassifierMixin, CVBaseEstimator):    
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            RandomForestClassifier()
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'randomforestclassifier__n_estimators': geom(1./100)
        })


class RandomForestRegressorCV(RegressorMixin, CVBaseEstimator):    
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            RandomForestRegressor()
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'randomforestregressor__n_estimators': geom(1./100)
        })


class PCARandomForestRegressorCV(RegressorMixin, CVBaseEstimator):    
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            RandomForestRegressor()
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'randomforestregressor__n_estimators': geom(1./100)
        })


class LogisticLassoCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['l1'],
            'logisticregression__solver': ['liblinear']
        })


class PCALogisticLassoCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['l1'],
            'logisticregression__solver': ['liblinear']
        })


class LassoLarsCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            LassoLars(normalize=True)
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'lassolars__alpha': expon(0, 1)
        })
    
    
class PCALassoLarsCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            LassoLars(normalize=True)
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'lassolars__alpha': expon(0, 1)
        })


class LogisticRidgeCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['l2']
        })


class PCALogisticRidgeCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['l2']
        })


class RidgeCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            Ridge(normalize=True)
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'ridge__alpha': expon(0, 1)
        })
    
    
class PCARidgeCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            Ridge(normalize=True)
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'ridge__alpha': expon(0, 1)
        })


class LogisticElasticNetCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['elasticnet'],
            'logisticregression__l1_ratio': uniform(0, 1),
            'logisticregression__solver': ['saga']
        })


class PCALogisticElasticNetCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            LogisticRegression()
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'logisticregression__C': expon(0, 1),
            'logisticregression__penalty': ['elasticnet'],
            'logisticregression__l1_ratio': uniform(0, 1),
            'logisticregression__solver': ['saga']
        })


class ElasticNetCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            ElasticNet(normalize=True)
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'elasticnet__alpha': expon(0, 1),
            'elasticnet__l1_ratio': uniform(0, 1)
        })
    
    
class PCAElasticNetCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            ElasticNet(normalize=True)
        ).set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'elasticnet__alpha': expon(0, 1),
            'elasticnet__l1_ratio': uniform(0, 1)
        })


class KernelRidgeCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            KernelRidge()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'kernelridge__alpha': expon(0, 1),
            'kernelridge__degree': geom(.5, loc=1),
            'kernelridge__kernel': ['linear', 'poly', 'rbf', 'laplacian']
        })
    
    
class PCAKernelRidgeCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            KernelRidge()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'kernelridge__alpha': expon(0, 1),
            'kernelridge__degree': geom(.5, loc=1),
            'kernelridge__kernel': ['linear', 'poly', 'rbf', 'laplacian']
        })


class SVCCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        return make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            SVC(probability=True)
        ).set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'svc__C': expon(0, 1),
            'svc__degree': geom(.3),
            'svc__kernel': ['linear', 'poly', 'rbf'],
        })
    
    
class PCASVCCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            SVC(probability=True)
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'svc__C': expon(0, 1),
            'svc__degree': geom(.3),
            'svc__kernel': ['linear', 'poly', 'rbf'],
        })


class SVRCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            SVR()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'svr__C': expon(0, 1),
            'svr__degree': geom(.3),
            'svr__kernel': ['linear', 'poly', 'rbf'],
        })
    
    
class PCASVRCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            SVR()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'svr__C': expon(0, 1),
            'svr__degree': geom(.3),
            'svr__kernel': ['linear', 'poly', 'rbf'],
        })

    
class KNeighborsClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            KNeighborsClassifier()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'kneighborsclassifier__n_neighbors': geom(1/(.05*X.shape[0])),
            'kneighborsclassifier__weights': ['uniform', 'distance']
        })


class PCAKNeighborsClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            KNeighborsClassifier()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'kneighborsclassifier__n_neighbors': geom(1/(.05*X.shape[0])),
            'kneighborsclassifier__weights': ['uniform', 'distance']
        })


class KNeighborsRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            StandardScaler(),
            KNeighborsRegressor()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'kneighborsregressor__n_neighbors': geom(1/(.05*X.shape[0])),
            'kneighborsregressor__weights': ['uniform', 'distance']
        })


class PCAKNeighborsRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            StandardScaler(),
            KNeighborsRegressor()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'kneighborsregressor__n_neighbors': geom(1/(.05*X.shape[0])),
            'kneighborsregressor__weights': ['uniform', 'distance']
        })
    

class AdaBoostClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            AdaBoostClassifier()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'adaboostclassifier__n_estimators': geom(1./2**5)
        })


class PCAAdaBoostClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            AdaBoostClassifier()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'adaboostclassifier__n_estimators': geom(1./2**5)
        })


class AdaBoostRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            AdaBoostRegressor()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'adaboostregressor__n_estimators': geom(1./2**5)
        })


class PCAAdaBoostRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            AdaBoostRegressor()
        )
        return est.set_params(**params)
        
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'adaboostregressor__n_estimators': geom(1./2**5)
        })


class XGBClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            XGBClassifier()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'xgbclassifier__gamma': expon(0, 1),
            'xgbclassifier__max_depth': geom(1./6),
            'xgbclassifier__min_child_weight': expon(0, 1),
            'xgbclassifier__max_delta_step': expon(0, 1),
            'xgbclassifier__lambda': expon(0, 1),
            'xgbclassifier__alpha': expon(0, 1),
        })


class PCAXGBClassifierCV(ClassifierMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            XGBClassifier()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'xgbclassifier__gamma': expon(0, 1),
            'xgbclassifier__max_depth': geom(1./6),
            'xgbclassifier__min_child_weight': expon(0, 1),
            'xgbclassifier__max_delta_step': expon(0, 1),
            'xgbclassifier__lambda': expon(0, 1),
            'xgbclassifier__alpha': expon(0, 1),
        })


class XGBRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            XGBRegressor()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'xgbregressor__gamma': expon(0, 1),
            'xgbregressor__max_depth': geom(1./6),
            'xgbregressor__min_child_weight': expon(0, 1),
            'xgbregressor__max_delta_step': expon(0, 1),
            'xgbregressor__lambda': expon(0, 1),
            'xgbregressor__alpha': expon(0, 1),
        })


class PCAXGBRegressorCV(RegressorMixin, CVBaseEstimator):
    def make_estimator(self, **params):
        est = make_pipeline(
            *self.preprocessors,
            PolynomialFeatures(),
            StandardScaler(),
            PCA(),
            XGBRegressor()
        )
        return est.set_params(**params)
    
    def get_param_distributions(self, X, y):
        return super().get_param_distributions({
            'polynomialfeatures__degree': [1, 2],
            'pca__n_components': list(range(1, X.shape[1])),
            'xgbregressor__gamma': expon(0, 1),
            'xgbregressor__max_depth': geom(1./6),
            'xgbregressor__min_child_weight': expon(0, 1),
            'xgbregressor__max_delta_step': expon(0, 1),
            'xgbregressor__lambda': expon(0, 1),
            'xgbregressor__alpha': expon(0, 1),
        })


"""# Linear models"""

import numpy as np
from sklearn.linear_model import LinearRegression, Ridge as RidgeBase


class ConstrainedLinearRegression(LinearRegression):
    """
    Linear regression where the coefficients are constrained to sum to a given 
    value.

    Parameters
    ----------
    constraint : scalar, default=0
        Sum of the regression coefficients.

    normalize : bool, default=False
    
    copy_Xbool, default=True
        If True, X will be copied; else, it may be overwritten.

    n_jobs : int, default=None
        The number of jobs to use for the computation. This will only provide speedup for n_targets > 1 and sufficient large problems. None means 1 unless in a joblib.parallel_backend context. -1 means using all processors.

    Attributes
    ----------
    coef_ : array-like of shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression contrained to sum to the given constraint value.

    Examples
    --------
    ```python
    from fast_automl.linear_model import ConstrainedLinearRegression

    from sklearn.datasets import load_boston
    from sklearn.model_selection import train_test_split

    X, y = load_boston(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True)
    reg = ConstrainedLinearRegression(constraint=.8).fit(X_train, y_train)
    print(reg.score(X_test, y_test))
    print(reg.coef_.sum())
    ```

    Out:

    ```
    0.6877629260102918
    0.8
    ```
    """
    def __init__(self, constraint=0, copy_X=True, n_jobs=None):
        self.constraint = constraint
        super().__init__(
            fit_intercept=False,
            copy_X=copy_X, 
            n_jobs=n_jobs
        )
        
    def fit(self, X, y, sample_weight=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        
        y : array-like of shape (n_samples, n_targets)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        if X.shape[1] == 1:
            self.coef_ = np.array([self.constraint])
            return self
        if hasattr(X, 'values'):
            X = X.values
        X_0, X_rest = X[:,0], X[:,1:]
        X_rest = (X_rest.T - X_0).T
        y = y - self.constraint * X_0
        super().fit(X_rest, y, sample_weight)
        self.coef_ = np.insert(
            self.coef_, 0, self.constraint - self.coef_.sum()
        )
        return self
    
    def predict(self, X):
        """
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples, n_targets)
            Predicted values
        """
        return X @ self.coef_


class Ridge(RidgeBase):
    """
    Ridge regression with the option for custom prior weights on coefficients.

    Parameters
    ----------
    prior_weight : array-like of shape (n_features)
        Prior weight means.

    See [scikit-learn's ridge regression documentation](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html) for additional parameter details.

    Attributes
    ----------
    coef_ndarray of shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    intercept_float or ndarray of shape (n_targets,)
        Independent term in decision function. Set to 0.0 if `fit_intercept = False`.

    Examples
    --------
    ```python
    from fast_automl.linear_model import Ridge

    from sklearn.datasets import load_boston
    from sklearn.model_selection import train_test_split

    X, y = load_boston(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True)
    reg = Ridge().fit(X_train, y_train)
    reg.score(X_test, y_test)
    ```
    """
    def __init__(
            self, alpha=1., prior_weight=0, normalize_coef=False, 
            fit_intercept=True, normalize=False, copy_X=True, max_iter=None, 
            tol=.001, solver='auto', random_state=None
        ):
        self.prior_weight = prior_weight
        self.normalize_coef = normalize_coef
        super().__init__(
            alpha, fit_intercept=fit_intercept, normalize=normalize, 
            copy_X=copy_X, max_iter=max_iter, tol=tol, solver=solver, 
            random_state=random_state
        )
        
    def fit(self, X, y, sample_weight=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        
        y : array-like of shape (n_samples, n_targets)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        y = y - (self.prior_weight * X).sum(axis=1)
        super().fit(X, y, sample_weight)
        if self.normalize_coef:
            self.coef_ -= self.coef_.mean()
        return self
    
    def predict(self, X):
        """
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples, n_targets)
            Predicted values
        """
        return super().predict(X) + (self.prior_weight * X).sum(axis=1)


"""# Utilities"""

import numpy as np
import pandas as pd
from sklearn.base import (
    BaseEstimator, RegressorMixin, 
    TransformerMixin as TransformerMixinBase
)


class TransformerMixin(TransformerMixinBase, BaseEstimator):
    """
    Version of scikit-learn's [`TransformerMixin`](https://scikit-learn.org/stable/modules/generated/sklearn.base.TransformerMixin.html) which implements a default inert `fit` method.
    """
    def fit(self, X, y=None):
        """
        This function doesn't do anything, but is necessary to include the transformer in a `Pipeline`.

        Returns
        -------
        self
        """
        return self
    
    def transform(self, X):
        """
        Must be implemented by the transformer.
        """
        raise NotImplementedError('Transfomer must implement transform method')


class ColumnSelector(TransformerMixin):
    """
    Selects columns from a dataframe.

    Parameters
    ----------
    columns : list
        List of columns to select.

    Examples
    --------
    ```python
    from fast_automl.utils import ColumnSelector

    import numpy as np
    import pandas as pd
    from sklearn.linear_model import LinearRegression
    from sklearn.pipeline import make_pipeline

    X = pd.DataFrame({
    \    'x0': [-1, -2, 1, 2],
    \    'x1': [-1, -1, 1, 1]
    })
    y = np.array([1, 1, 2, 2])

    reg = make_pipeline(
    \    ColumnSelector(['x1']),
    \    LinearRegression()
    ).fit(X, y)
    reg.score(X, y)
    ```
    """
    def __init__(self, columns):
        self.columns = columns

    def transform(self, X, y=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        
        y : optional, array-like of shape (n_samples, n_targets)
            Target values.

        Returns
        -------
        X or (X, y) : Where X columns have been selected
        """
        X = (
            X[self.columns] if isinstance(X, pd.DataFrame) 
            else X[:,self.columns]
        )
        return X if y is None else (X, y)


class ColumnRemover(TransformerMixin):
    """
    Removes columns from a dataframe.

    Parameters
    ----------
    columns : list
        List of columns to remove.

    Examples
    --------
    ```python
    from fast_automl.utils import ColumnRemover

    import numpy as np
    import pandas as pd
    from sklearn.linear_model import LinearRegression
    from sklearn.pipeline import make_pipeline

    X = pd.DataFrame({
    \    'x0': [-1, -2, 1, 2],
    \    'x1': [-1, -1, 1, 1]
    })
    y = np.array([1, 1, 2, 2])

    reg = make_pipeline(
    \    ColumnRemover(['x0']),
    \    LinearRegression()
    ).fit(X, y)
    reg.score(X, y)
    ```
    """
    def __init__(self, columns):
        self.columns = columns

    def transform(self, X, y=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        
        y : optional, array-like of shape (n_samples, n_targets)
            Target values.

        Returns
        -------
        X or (X, y) : Where X columns have been removed
        """
        X = (
            X.drop(columns=self.columns) if isinstance(X, pd.DataFrame)
            else np.delete(X, self.columns, axis=1)
        )
        return X if y is None else (X, y)


class BoundRegressor(RegressorMixin, BaseEstimator):
    """
    Constrains the predicted target value to be within the range of targets in
    the training data.

    Parameters
    ----------
    estimator : scikit-learn style regressor

    Attributes
    ----------
    estimator_ : scikit-learn style regressor
        Fitted regressor.

    y_max_ : scalar
        Maximum target value in training data.

    y_min_ : scalar
        Minimum target value in training data.

    Examples
    --------
    ```python
    from fast_automl.utils import BoundRegressor

    import numpy as np
    from sklearn.linear_model import LinearRegression

    X_train = np.array([
    \    [1, 2],
    \    [7, 8]
    ])
    X_test = np.array([
    \    [3, 4],
    \    [5, 1000]
    ])
    y_train = np.array([1.5, 7.5])
    y_test = np.array([3.5, 5.5])

    reg = LinearRegression().fit(X_train, y_train)
    reg.predict(X_test)
    ```

    Out:

    ```
    array([3.5, 7.5])
    ```
    """
    def __init__(self, estimator):
        self.estimator = estimator
        
    def fit(self, X, y, sample_weight=None):
        self.estimator_ = self.estimator.fit(X, y, sample_weight)
        self.y_min_, self.y_max_ = y.min(), y.max()
        return self
    
    def predict(self, X):
        y_pred = self.estimator_.predict(X)
        return np.clip(y_pred, self.y_min_, self.y_max_)


"""# Ensemble"""

from .linear_model import ConstrainedLinearRegression
from .metrics import check_scoring

import numpy as np
from joblib import Parallel
from scipy import optimize
from scipy.stats import loguniform
from sklearn.base import (
    BaseEstimator, ClassifierMixin, RegressorMixin, clone, is_classifier
)
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.ensemble._base import _fit_single_estimator
from sklearn.ensemble._stacking import _BaseStacking
from sklearn.metrics import log_loss, roc_auc_score
from sklearn.model_selection import (
    check_cv, cross_val_predict, cross_val_score
)
from sklearn.utils.fixes import delayed

from copy import deepcopy

def _predict_single_estimator(estimator, X, method='predict'):
    assert method in ('predict', 'predict_proba')
    return getattr(estimator, method)(X)


class ClassifierWeighter(ClassifierMixin, BaseEstimator):
    """
    Trains weights for ensemble of classifiers.

    Parameters
    ----------
    loss : callable, default=log_loss
        Loss function to minimize by weight fitting.

    Attributes
    ----------
    coef_ : array-like of shape (n_estimators,)
        Weights on the given estimators.

    Examples
    --------
    ```python
    from fast_automl.ensemble import ClassifierWeighter

    import numpy as np
    from sklearn.datasets import load_breast_cancer
    from sklearn.model_selection import StratifiedKFold, cross_val_predict, train_test_split
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.svm import SVC

    from copy import deepcopy

    X, y = load_breast_cancer(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, shuffle=True)

    svc = SVC(probability=True).fit(X_train, y_train)
    knn = KNeighborsClassifier().fit(X_train, y_train)

    cv = StratifiedKFold(random_state=np.random.RandomState(), shuffle=True)
    X_meta = np.array([
    \    cross_val_predict(clf, X_train, y_train, cv=deepcopy(cv), method='predict_proba')
    \    for clf in (svc, knn)
    ]).transpose(1, 2, 0)
    weighter = ClassifierWeighter().fit(X_meta, y_train)

    X_meta_test = np.array([
    \    clf.predict_proba(X_test) for clf in (svc, knn)
    ]).transpose(1, 2, 0)
    weighter.score(X_meta_test, y_test)
    ```
    """
    def __init__(self, loss=log_loss):
        self.loss = loss
        
    def fit(self, X, y):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_classes, n_estimators)
            X_ice is the probability estimator e puts on sample i being in 
            class c.

        y : array-like of shape (n_samples,)
            Targets.

        Returns
        -------
        self
        """
        # X is (n_samples, n_classes, n_estimators)
        def loss(coef_):
            return self.loss(y, (X*coef_).sum(axis=-1))
        
        n_estimators = X.shape[-1]
        self.classes_ = np.arange(X.shape[1])
        if n_estimators == 1:
            self.coef_ = np.array([1])
            return self
        constraint = optimize.LinearConstraint(np.array([1]*n_estimators), 1, 1)
        coef0 = np.array([1./n_estimators]*n_estimators)
        self.coef_ = optimize.minimize(loss, x0=coef0, constraints=[constraint]).x
        return self

    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array of shape (n_samples,)
            Predicted class label for each sample.
        """
        return np.argmax(self.predict_proba(X), axis=1)
    
    def predict_proba(self, X):
        """
        Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the sample for each classes.
        """
        return (X*self.coef_).sum(axis=-1)


class BaseStackingCV(_BaseStacking):
    """
    Parameters
    ----------
    estimators : list
        Base estimators which will be stacked together. Each element of the 
        list is defined as a tuple of string (i.e. name) and an estimator 
        instance.

    cv : int, cross-validation generator, or iterable, default=None
        Scikit-learn style cv parameter.

    shuffle_cv : bool, default=True
        Indicates that cross validator should shuffle observations.

    scoring : str, callable, list, tuple, or dict, default=None
        Scikit-learn style scoring parameter.

    n_jobs : int, default=None
        Number of jobs to run in parallel.

    verbose : int, default=0
        Controls the verbosity.
    """
    def __init__(
            self, estimators, cv=None, shuffle_cv=True, scoring=None, 
            n_jobs=None, verbose=0
        ):
        self.shuffle_cv = shuffle_cv
        self.scoring = scoring
        super().__init__(estimators, cv=cv, n_jobs=n_jobs, verbose=verbose)

    def fit(self, X, y, sample_weight=None):
        raise NotImplementedError()

    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array of shape (n_samples,)
            Predicted outcome for each sample.
        """
        assert hasattr(self, 'best_estimator_'), (
            'Estimator has not been fitted. Call `fit` before `predict`'
        )
        return self.best_estimator_.predict(X)
    
    def predict_proba(self, X):
        """
        Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the sample for each classes on the model.

        Notes
        -----
        Only applicable to classifiers.
        """
        assert hasattr(self, 'best_estimator_'), (
            'Estimator has not been fitted. Call `fit` before `predict`'
        )
        return self.best_estimator_.predict_proba(X)
        
    def transform(self, X):
        """
        Transforms raw features into a prediction matrix.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Features.

        Returns
        -------
        X_meta : array-like
            Prediction matrix of shape (n_samples, n_estimators) for 
            regression and (n_estimators, n_samples, n_classes) for 
            classification.
        """
        predictions = Parallel(n_jobs=self.n_jobs)(
            delayed(_predict_single_estimator)(est, X)
            for est in self.estimators_
        )
        return self._concatenate_predictions(X, predictions)
    
    def _fit_estimators(self, X, y, estimators, sample_weight):
        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_single_estimator)(clone(est), X, y, sample_weight)
            for _, est in estimators
        )
        
    def _check_cv(self, y):
        cv = check_cv(self.cv, y=y, classifier=is_classifier(self))
        if hasattr(cv, 'random_state') and cv.random_state is None:
            cv.random_state = np.random.RandomState()
        if hasattr(cv, 'shuffle') and self.shuffle_cv:
            cv.shuffle = True
        return cv
    
    def _cross_val_predict(self, X, y, estimators, cv, sample_weight):
        fit_params = (
            {} if sample_weight is None else {'sample_weight': sample_weight}
        )
        method = 'predict_proba' if is_classifier(self) else 'predict'
        predictions = Parallel(n_jobs=self.n_jobs)(
            delayed(cross_val_predict)(
                clone(est), X, y, 
                cv=deepcopy(cv), 
                n_jobs=self.n_jobs, 
                fit_params=fit_params, 
                verbose=self.verbose,
                method=method
            )
            for est in estimators
        )
        if is_classifier(self):
            return np.array(predictions).transpose(1, 2, 0)
        return np.array(predictions).T
    
    def _make_best_estimator(self, estimators, weights):
        if is_classifier(self):
            voting_estimator_cls = VotingClassifier
            kwargs = {'voting': 'soft'}
        else:
            voting_estimator_cls = VotingRegressor
            kwargs = {}
        return voting_estimator_cls(
            estimators, weights=weights,
            n_jobs=self.n_jobs, verbose=self.verbose, **kwargs
        )
    

class RFEVotingEstimatorCV(BaseStackingCV):
    """
    Selects estimators using recursive feature elimination. Inherits from 
    `BaseStackingCV`.

    Attributes
    ----------
    best_estimator_ : estimator
        The voting estimator associated with the highest CV score.

    best_score_ : scalar
        Highest CV score attained by any voting estimator.

    weights_ : array-like
        Weights the voting estimator places on each of the estimators in its 
        ensemble.

    names_ : list
        List of estimator names in the best estimator.
    """
    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        def compute_rfe_scores(X_meta):
            # compute CV scores for each step of recursive feature elimination
            estimators = self.estimators.copy()
            rfe_scores = []
            while estimators:
                weight_fitter.fit(X_meta, y)
                # if any weights are negative, estimator has overfit
                # assign -inf as the score
                if np.any(weight_fitter.coef_ < 0):
                    score = -np.inf
                else:
                    score = cross_val_score(
                        weight_fitter, X_meta, y, cv=cv, scoring=scoring
                    ).mean()
                # append score, current estimators, weights
                rfe_scores.append(
                    (score, estimators.copy(), weight_fitter.coef_)
                )
                # drop estimator with the least weight
                drop_idx = int(np.argmin(weight_fitter.coef_))
                estimators.pop(drop_idx)
                X_meta = np.delete(X_meta, drop_idx, axis=-1)
            return max(rfe_scores, key=lambda x: x[0])
        
        if is_classifier(self):
            weight_fitter = ClassifierWeighter()
            best_estimator_cls = VotingClassifier
        else:
            weight_fitter = ConstrainedLinearRegression(1)
            best_estimator_cls = VotingRegressor
        names, all_estimators = self._validate_estimators()
        cv = self._check_cv(y)
        scoring = check_scoring(self.scoring, classifier=is_classifier(self))
        X_meta = self._cross_val_predict(
            X, y, all_estimators, cv, sample_weight
        )
        self.best_score_, estimators, self.weights_ = compute_rfe_scores(X_meta)
        self.names_ = [name for name, _ in estimators]
        self._fit_estimators(X, y, estimators, sample_weight)
        self.best_estimator_ = self._make_best_estimator(
            estimators, self.weights_
        ).fit(X, y)
        return self


class RFEVotingClassifierCV(ClassifierMixin, RFEVotingEstimatorCV):
    """
    Selects classifiers using recursive feature elimination. Inherits from 
    `RFEVotingEstimatorCV`.

    Examples
    --------
    ```python
    from fast_automl.ensemble import RFEVotingClassifierCV

    from sklearn.datasets import load_breast_cancer
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import cross_val_score, train_test_split
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.svm import SVC

    X, y = load_breast_cancer(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, shuffle=True)

    clf = RFEVotingClassifierCV([
    \    ('rf', RandomForestClassifier()),
    \    ('knn', KNeighborsClassifier()),
    \    ('svm', SVC(probability=True))
    ]).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(clf.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(clf.score(X_test, y_test)))
    ```
    """


class RFEVotingRegressorCV(RegressorMixin, RFEVotingEstimatorCV):
    """
    Selects regressors using recursive feature elimination. Inherits from 
    `RFEVotingEstimatorCV`.

    Examples
    --------
    ```python
    from fast_automl.ensemble import RFEVotingRegressorCV

    from sklearn.datasets import load_boston
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.model_selection import cross_val_score, train_test_split
    from sklearn.neighbors import KNeighborsRegressor
    from sklearn.svm import SVR

    X, y = load_boston(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True)

    reg = RFEVotingRegressorCV([
    \    ('rf', RandomForestRegressor()),
    \    ('knn', KNeighborsRegressor()),
    \    ('svm', SVR())
    ]).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(reg.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(reg.score(X_test, y_test)))
    ```
    """


class StepwiseVotingEstimatorCV(BaseStackingCV):
    """
    Selects estimators using stepwise addition. Inherits from 
    `BaseStackingCV`.

    Attributes
    ----------
    best_estimator_ : estimator
        The voting estimator associated with the highest CV score.

    best_score_ : scalar
        Highest CV score attained by any voting estimator.

    weights_ : array-like
        Weights the voting estimator places on each of the estimators in its 
        ensemble.

    names_ : list
        List of estimator names in the best estimator.
    """
    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        def compute_stepwise_scores():
            # compute CV score from stepwise addition of estimators to the 
            # ensemble
            best_score = -np.inf
            # estimators in the ensemble
            in_estimators, X_in = [], None
            # estimators out of the ensemble
            out_estimators, X_out = self.estimators.copy(), X_meta
            while out_estimators:
                # add estimators from outside the ensemble to the ensemble
                new_scores = [compute_new_score(X_in, col) for col in X_out.T]
                idx = np.argmax(new_scores)
                if new_scores[idx] <= best_score:
                    break
                # if the new estimator improved the CV score
                # add it to the ensemble
                best_score = new_scores[idx]
                in_estimators.append(out_estimators.pop(idx))
                if is_classifier(self):
                    col = X_out[:, :, idx].reshape(X_meta.shape[0], X_meta.shape[1], 1)
                else:
                    col = X_out[:, idx].reshape(-1, 1)
                X_in = (
                    col if X_in is None 
                    else np.concatenate((X_in, col), axis=-1)
                )
                X_out = np.delete(X_out, idx, axis=-1)
            return best_score, in_estimators, weight_fitter.fit(X_in, y).coef_
        
        def compute_new_score(X_in, col):
            # add new estimator CV predictions to the dataframe
            if is_classifier(self):
                col = col.T.reshape(X_meta.shape[0], X_meta.shape[1], 1)
            else:
                col = col.reshape(-1, 1)
            X = (
                col if X_in is None 
                else np.concatenate((X_in, col), axis=-1)
            )
            weights = weight_fitter.fit(X, y).coef_
            # add the CV score for the estimator to new_scores
            # if any of the weights on the estimators are negative, 
            # the ensemble has overfit; assign a weight of -inf
            if np.any(weights < 0):
                return -np.inf
            return cross_val_score(
                weight_fitter, X, y, cv=cv, scoring=scoring
            ).mean()
            
        if is_classifier(self):
            weight_fitter = ClassifierWeighter()
            best_estimator_cls = VotingClassifier
            best_estimator_kwargs = {'voting': 'soft'}
        else:
            weight_fitter = ConstrainedLinearRegression(1)
            best_estimator_cls = VotingRegressor
            best_estimator_kwargs = {}
        names, all_estimators = self._validate_estimators()
        cv = self._check_cv(y)
        scoring = check_scoring(self.scoring, classifier=is_classifier(self))
        X_meta = self._cross_val_predict(
            X, y, all_estimators, cv, sample_weight
        )
        self.best_score_, estimators, self.weights_ = compute_stepwise_scores()
        self.names_ = [name for name, _ in estimators]
        self._fit_estimators(X, y, estimators, sample_weight)
        self.best_estimator_ = self._make_best_estimator(
            estimators, self.weights_
        ).fit(X, y)
        return self
    

class StepwiseVotingClassifierCV(ClassifierMixin, StepwiseVotingEstimatorCV):
    """
    Selects classifiers using stepwise addition. Inherits from 
    `StepwiseVotingEstimatorCV`.

    Examples
    --------
    ```python
    from fast_automl.ensemble import StepwiseVotingClassifierCV

    from sklearn.datasets import load_iris
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import cross_val_score, train_test_split
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.svm import SVC

    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, shuffle=True)

    clf = StepwiseVotingClassifierCV([
    \    ('rf', RandomForestClassifier()),
    \    ('knn', KNeighborsClassifier()),
    \    ('svm', SVC(probability=True))
    ]).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(clf.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(clf.score(X_test, y_test)))
    ```
    """

    
class StepwiseVotingRegressorCV(RegressorMixin, StepwiseVotingEstimatorCV):
    """
    Selects regressors using stepwise addition. Inherits from 
    `StepwiseVotingEstimatorCV`.

    Examples
    --------
    ```python
    from fast_automl.ensemble import StepwiseVotingRegressorCV

    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.model_selection import cross_val_score, train_test_split
    from sklearn.neighbors import KNeighborsRegressor
    from sklearn.svm import SVR

    X, y = load_diabetes(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True)

    reg = StepwiseVotingRegressorCV([
    \    ('rf', RandomForestRegressor()),
    \    ('knn', KNeighborsRegressor()),
    \    ('svm', SVR())
    ]).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(reg.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(reg.score(X_test, y_test)))
    ```
    """


"""# Metrics"""

from sklearn.metrics import (
    r2_score, roc_auc_score as roc_auc_score_base, make_scorer
)

def check_scoring(scoring, classifier):
    """
    Creates a default regression or classifier scoring rule. This is R-squared for regressors and ROC AUC for classifiers.

    Parameters
    ----------
    scoring : str or callable
        A str (see scikit-learn's model evaluation docs) or a scorer callable with signature `scorer(estimator, X, y)` which returns a single value.

    classifier : bool
        Indicates that the estimator is a classifier.
    """
    if scoring is not None:
        return scoring
    return roc_auc_scorer if classifier else make_scorer(r2_score)

def roc_auc_score(y, output):
    """
    Parameters
    ----------
    y : array-like of shape (n_samples,)
        target values.

    output : array-like of shape (n_samples, n_classes)
        Predicted probability of each class.

    Returns
    -------
    score : scalar
        ROC AUC score, default is one-versus-rest for multi-class problems.    
    """
    if len(output.shape) == 1:
        # binary classifier
        return roc_auc_score_base(y, output)
    # multi-class
    return roc_auc_score_base(y, output, multi_class='ovr')

roc_auc_scorer = make_scorer(roc_auc_score, needs_proba=True)


"""# Baseline classifier and regressor"""

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import _check_sample_weight


class BaselineClassifier(ClassifierMixin, BaseEstimator):
    """
    Predicts the most frequent class.

    Attributes
    ----------
    classes_ : array-like of shape (n_classes,)
        A list of class weights known to the classifier.

    counts_ : array-like of shape (n_classes,)
        Normalized frequency of each class in the training data.

    dominant_class_ : int
        Class which appears most frequently in the training data.

    Examples
    --------
    ```python
    from fast_automl.baseline import BaselineClassifier

    from sklearn.datasets import load_digits
    from sklearn.model_selection import train_test_split

    X, y = load_digits(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y)
    clf = BaselineClassifier().fit(X_train, y_train)
    clf.score(X_test, y_test)
    ```
    """
    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        check_classification_targets(y)
        if sample_weight is None:
            self.classes_, self.counts_ = np.unique(y, return_counts=True)
        else:
            sample_weight = _check_sample_weight(sample_weight, X)
            sample_weight = sample_weight / sample_weight.mean()
            df = pd.DataFrame({'y': y, 'sample_weight': sample_weight})
            df = df.groupby('y').sum()
            self.classes_ = df.index.values
            self.counts_ = df.sample_weight.values
        self.counts_ = self.counts_ / self.counts_.sum()
        self.dominant_class_ = self.classes_[np.argmax(self.counts_)]
        return self
    
    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array of shape (n_samples,)
            Predicted class label for each sample.
        """
        return np.array([self.dominant_class_]*X.shape[0])
    
    def predict_proba(self, X):
        """
        Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the sample for each classes on the model, ordered by `self.classes_`.
        """
        return np.array([self.counts_]*X.shape[0])


class BaselineRegressor(RegressorMixin, BaseEstimator):
    """
    Predicts the mean target value.

    Attributes
    ----------
    y_mean_ : np.array
        Average target value.

    Examples
    --------
    ```python
    from fast_automl.baseline import BaselineRegressor

    from sklearn.datasets import load_boston
    from sklearn.model_selection import train_test_split

    X, y = load_boston(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y)
    reg = BaselineRegressor().fit(X_train, y_train)
    reg.score(X_test, y_test)
    ```
    """
    def fit(self, X, y, sample_weight=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        
        y : array-like of shape (n_samples, n_targets)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        sample_weight = _check_sample_weight(sample_weight, X)
        self.y_mean_ = (y * sample_weight).mean() / sample_weight.mean()
        return self
        
    def predict(self, X):
        """
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples, n_targets)
            Predicted values
        """
        return self.y_mean_ * np.ones(shape=X.shape[0])


"""# Automated machine learning"""

from .cv_estimators import *
from .ensemble import (
    RFEVotingClassifierCV, RFEVotingRegressorCV, 
    StepwiseVotingClassifierCV, StepwiseVotingRegressorCV
)
from .metrics import check_scoring

import numpy as np
from sklearn.base import (
    BaseEstimator, ClassifierMixin, RegressorMixin, is_classifier
)
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import make_pipeline

def make_cv_regressors():
    """
    Returns
    -------
    cv_regressors : list
        List of default CV regresssors.
    """
    return [
        RandomForestRegressorCV(),
        PCARandomForestRegressorCV(),
        LassoLarsCV(),
        PCALassoLarsCV(),
        RidgeCV(),
        PCARidgeCV(),
        ElasticNetCV(),
        PCAElasticNetCV(),
        KernelRidgeCV(),
        PCAKernelRidgeCV(),
        SVRCV(),
        PCASVRCV(),
        KNeighborsRegressorCV(),
        PCAKNeighborsRegressorCV(),
        AdaBoostRegressorCV(),
        PCAAdaBoostRegressorCV(),
        XGBRegressorCV(),
        PCAXGBRegressorCV()
    ]


def make_cv_classifiers():
    """
    Returns
    -------
    cv_classifiers : list
        List of default CV classifiers.
    """
    return [
        RandomForestClassifierCV(),
        PCARandomForestClassifierCV(),
        LogisticLassoCV(),
        PCALogisticLassoCV(),
        LogisticRidgeCV(),
        PCALogisticRidgeCV(),
        LogisticElasticNetCV(),
        PCALogisticElasticNetCV(),
        SVCCV(),
        PCASVCCV(),
        KNeighborsClassifierCV(),
        PCAKNeighborsClassifierCV(),
        AdaBoostClassifierCV(),
        PCAAdaBoostClassifierCV(),
        XGBClassifierCV(),
        PCAXGBClassifierCV()
    ]

class AutoEstimator(BaseEstimator):
    """
    Parameters
    ----------
    cv_estimators : list of CVEstimators, default=[]
        If an empty list, a default list of CVEstimators will be created.

    preprocessors : list, default=[]
        List of preprocessing steps before data is fed to the `cv_estimators`.

    ensemble_method : str, default='auto'
        If `'rfe'`, the ensemble is created using recursive feature 
        elimination. If `'stepwise'`, the ensemble is created using stepwise 
        addition. If `'auto'`, the ensemble is the better of the RFE and 
        stepwise ensemble methods.

    max_ensemble_size : int, default=50
        The maximum number of estimators to consider adding to the ensemble.

    n_ensembles : int, default=1
        Number of ensembles to create using different CV splits. These 
        ensembles get equal votes in a meta-ensemble.

    n_iter : int, default=10
        Number of iterations to run randomized search for the CVEstimators.

    n_jobs : int or None, default=None
        Number of jobs to run in parallel.

    verbose : bool, default=False
        Controls the verbosity.

    cv : int, cross-validation generator, or iterable, default=None
        Scikit-learn style cv parameter.

    scoring : str, callable, list, tuple, or dict, default=None
        Scikit-learn style scoring parameter. By default, a regressor 
        ensembles maximizes R-squared and a classifier ensemble maximizes ROC 
        AUC.

    Attributes
    ----------
    best_estimator_ : estimator
        Ensemble or meta-ensemble associated with the best CV score.
    """
    def __init__(
            self, cv_estimators=[], preprocessors=[], ensemble_method='auto', 
            max_ensemble_size=50, n_ensembles=1, n_iter=10, n_jobs=None, 
            verbose=False, cv=None, scoring=None
        ):
        if cv_estimators:
            self.cv_estimators = cv_estimators
        else:
            self.cv_estimators = (
                make_cv_classifiers() if is_classifier(self)
                else make_cv_regressors()
            )
        self.preprocessors = (
            preprocessors if isinstance(preprocessors, list) 
            else [preprocessors]
        )
        assert ensemble_method in ('auto', 'rfe', 'stepwise')
        self.ensemble_method = ensemble_method
        self.max_ensemble_size = max_ensemble_size
        self.n_ensembles = n_ensembles
        self.n_iter = n_iter
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.cv = cv
        self.scoring = scoring
        
    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        sample_weight, array-like of shape (n_samples,), default=Noone
            Individual weights for each sample.

        Returns
        -------
        self
        """
        def fit_cv_estimator(i, est):
            if self.verbose:
                print('\nTuning estimator {} of {}: {}'.format(
                    i+1, len(self.cv_estimators), est.__class__.__name__
                ))
            est.fit(
                X, y, n_iter=self.n_iter, n_jobs=self.n_jobs, scoring=scoring
            )
            if self.verbose:
                print('Best estimator score: {:.4f}'.format(est.best_score_))

        def make_best_estimators():
            cv_results = [
                (res, est) 
                for est in self.cv_estimators 
                for res in est.cv_results_
            ]
            # create the best estimators
            # x[0][0] is the cv score
            best_params = sorted(
                cv_results, key=lambda x: x[0][0], reverse=True
            )[:self.max_ensemble_size]
            best_estimators = [
                est.make_estimator(**params) 
                for (_, params), est in best_params
            ]
            return [
                ('estimator {}'.format(i), estimator) 
                for i, estimator in enumerate(best_estimators)
            ]

        def build_ensemble(i):
            # build an ensemble of the best estimators
            if self.verbose:
                print(
                    '\nBuilding ensemble {} of {}'.format(
                        i+1, self.n_ensembles
                    )
                )
            ensemble_classes = []
            if self.ensemble_method in ('rfe', 'auto'):
                if is_classifier(self):
                    ensemble_classes.append(RFEVotingClassifierCV)
                else:
                    ensemble_classes.append(RFEVotingRegressorCV)
            elif self.ensemble_method in ('stepwise', 'auto'):
                if is_classifier(self):
                    ensemble_classes.append(StepwiseVotingClassifierCV)
                else:
                    ensemble_classes.append(StepwiseVotingRegressorCV)
            # TODO when ensemble_method is 'auto', both the RFE and stepwise estimators will have their sub-estimators make CV predictions, which is redundant work. A future release should get a CV split and make CV predictions here, then pass these to stepwise and RFE selection methods
            ensembles = [
                ensemble_cls(
                    best_estimators, n_jobs=self.n_jobs, cv=self.cv, scoring=scoring
                ).fit(X, y, sample_weight=sample_weight)
                for ensemble_cls in ensemble_classes
            ]
            ensemble = max(ensembles, key=lambda e: e.best_score_)
            if self.verbose:
                print('Best ensemble score: {:.4f}'.format(
                    ensemble.best_score_
                ))
            return 'ensemble {}'.format(i+1), ensemble.best_estimator_

        # store a copy of X to fit the best_estimator_
        X_copy = X.copy()
        for preprocessor in self.preprocessors:
            X = preprocessor.fit_transform(X)
        scoring = check_scoring(self.scoring, classifier=is_classifier(self))
        [fit_cv_estimator(i, est) for i, est in enumerate(self.cv_estimators)]
        best_estimators = make_best_estimators()
        ensembles = [
            build_ensemble(i) for i in range(self.n_ensembles)
        ]
        # if only one ensemble is built, there's no need to create a 
        # meta-ensemble
        # if multiple ensembles are built, give them an equal vote in a voting 
        # estimator
        if len(ensembles) == 1:
            meta_ensemble = ensembles[0][-1]
        elif is_classifier(self):
            meta_ensemble = VotingClassifier(ensembles, voting='soft')
        else:
            meta_ensemble = VotingRegressor(ensembles)
        self.best_estimator_ = make_pipeline(
            *self.preprocessors, meta_ensemble
        ).fit(X_copy, y)
        return self
    
    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array of shape (n_samples,)
            Predicted class label for each sample.
        """
        return self.best_estimator_.predict(X)
    
    def predict_proba(self, X):
        """
        Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the sample for each classes on the model, ordered by `self.classes_`.
        """
        return self.best_estimator_.predict_proba(X)


class AutoClassifier(ClassifierMixin, AutoEstimator):
    """
    Automatic classifier. Inherits from `AutoEstimator`.

    Examples
    --------
    ```python
    from fast_automl.automl import AutoClassifier

    from sklearn.datasets import load_digits
    from sklearn.model_selection import cross_val_score, train_test_split

    X, y = load_digits(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True, stratify=y)

    clf = AutoClassifier(ensemble_method='stepwise', n_jobs=-1, verbose=True).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(clf.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(clf.score(X_test, y_test)))
    ```

    This runs for about 6-7 minutes and typically achieves a test accuracy of 
    96-99% and ROC AUC above .999.
    """


class AutoRegressor(RegressorMixin, AutoEstimator):
    """
    Automatic regressor. Inherits from `AutoEstimator`.

    Examples
    --------
    ```python
    from fast_automl.automl import AutoRegressor

    from sklearn.datasets import load_diabetes
    from sklearn.model_selection import cross_val_score, train_test_split

    X, y = load_diabetes(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=True)

    reg = AutoRegressor(n_jobs=-1, verbose=True).fit(X_train, y_train)
    print('CV score: {:.4f}'.format(cross_val_score(reg.best_estimator_, X_train, y_train).mean()))
    print('Test score: {:.4f}'.format(reg.score(X_test, y_test)))
    ```

    This runs for about 30 seconds and typically achieves a test R-squared of 
    .47-.53.
    """